import os
from dataclasses import dataclass
from itertools import combinations
from typing import Optional, Tuple, Union

import numpy as np
import torch
from scipy import spatial
from scipy.optimize import Bounds
from scipy.optimize import minimize as scipy_minimize
from scipy.stats import norm

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.acquisition_functions.base_acq_function import AcquisitionFunction
from nemo_bo.acquisition_functions.expected_improvement.ehvi import ExpectedHypervolumeImprovement
from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList
from nemo_bo.opt.samplers import PoolBased, SampleGenerator, LatinHyperCubeSampling
from nemo_bo.opt.variables import VariablesList
from nemo_bo.utils.data_proc import remove_all_nan_rows

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass
class ExpectedImprovement(AcquisitionFunction):
    """

    Class to instantiate to use the acquisition functions expected improvement for single-objective problems or
    expected hypervolume improvement for multi-objective problems

    Parameters
    ----------
    num_candidates: int
        The number of sets of X arrays to be suggested by the acquisition function

    """

    num_candidates: int

    def generate_candidates(
        self,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        sampler: Union[SampleGenerator, PoolBased],
        constraints: Optional[ConstraintsList] = None,
        **kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that is called when generate the best candidates using the expected improvement or expected
        hypervolume acquisition functions. This selection is made automatically based on the number of objectives

        Parameters
        ----------
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        sampler: SampleGenerator | PoolBased
            Used during the ei or ehvi functions to generate samples (SampleGenerator) or provide the samples defined
            by the user (PoolBased)
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest expected improvement or expected
            hypervolume improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest expected improvement or expected
            hypervolume improvement

        """
        self.sampler = sampler

        Y = remove_all_nan_rows(Y)

        # if isinstance(self.sampler, SampleGenerator):
        #     self.sampler.num_new_points = kwargs.get("num_new_points")
        #     if self.sampler.num_new_points is None:
        #         # num_new_points = 5000
        #         if self.num_candidates == 1:
        #             self.sampler.num_new_points = 2000000
        #             # self.sampler.num_new_points = 500000
        #             # self.sampler.num_new_points = 500
        #         elif self.num_candidates == 2:
        #             self.sampler.num_new_points = 1000
        #             # self.sampler.num_new_points = 50
        #         elif self.num_candidates == 3:
        #             self.sampler.num_new_points = 100
        #         elif self.num_candidates == 4:
        #             # self.sampler.num_new_points = 100 # approx 4 million combinations
        #             self.sampler.num_new_points = 75  # approx 1.2 million combinations
        #         elif self.num_candidates == 5:
        #             # self.sampler.num_new_points = 60 # approx 5.5 million combinations
        #             self.sampler.num_new_points = 15
        # elif isinstance(self.sampler, PoolBased):
        if isinstance(self.sampler, PoolBased):
            # self.sampler.num_new_points = 1
            if self.num_candidates > 4:
                logger.warning(
                    f"The number of suggested candidates is 5 or greater, which is not recommended for pool-based samples with more than 60 samples due to long run times"
                )

        if objectives.n_obj > 1:
            if isinstance(self.sampler, SampleGenerator):
                # self.sampler.num_new_points = kwargs.get("num_new_points")
                if self.sampler.num_new_points is None:
                    # num_new_points = 5000
                    if self.num_candidates == 1:
                        self.sampler.num_new_points = 500000
                        # self.sampler.num_new_points = 500
                    elif self.num_candidates == 2:
                        self.sampler.num_new_points = 1000
                        # self.sampler.num_new_points = 50
                    elif self.num_candidates == 3:
                        self.sampler.num_new_points = 100
                    elif self.num_candidates == 4:
                        # self.sampler.num_new_points = 100 # approx 4 million combinations
                        self.sampler.num_new_points = 75  # approx 1.2 million combinations
                    elif self.num_candidates == 5:
                        # self.sampler.num_new_points = 60 # approx 5.5 million combinations
                        self.sampler.num_new_points = 15
            selected_X, selected_Y = self.ehvi(Y, variables, objectives, constraints)

        elif objectives.n_obj == 1:
            if isinstance(self.sampler, SampleGenerator):
                # self.sampler.num_new_points = kwargs.get("num_new_points")
                if self.sampler.num_new_points is None:
                    if isinstance(self.sampler, LatinHyperCubeSampling):
                        self.sampler.num_new_points = 2000000
                    else:
                        self.sampler.num_new_points = 1024

            self.xi = kwargs.get("xi", 0.01)
            selected_X, selected_Y = self.ei(Y, variables, objectives, constraints)

        self.reset = False

        return selected_X, selected_Y

    def ei(
        self,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        constraints: Optional[ConstraintsList] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that determines the candidates with the best expected improvement

        Parameters
        ----------
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest expected improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest expected improvement

        """
        logger.info(f"Identifying conditions with the highest expected hypervolume improvement")

        if isinstance(self.sampler, PoolBased):
            logger.info(f"Returning X values from the provided sample pool")
            X_new_points = self.sampler.X_pool
        elif isinstance(self.sampler, SampleGenerator):
            X_new_points = self.sampler.generate_samples(variables, constraints)

        Y_new_points, Y_new_points_stddev = objectives.evaluate(X_new_points)

        Y_norm = (Y - objectives.bounds[0, 0]) / (objectives.bounds[0, 1] - objectives.bounds[0, 0])
        Y_new_points_norm = (Y_new_points - objectives.bounds[0, 0]) / (
            objectives.bounds[0, 1] - objectives.bounds[0, 0]
        )
        # Min_max normalisation of Y standard deviation
        # I think this should be fine to minmax normalise like this
        Y_new_points_stddev_norm = Y_new_points_stddev / (objectives.bounds[0, 1] - objectives.bounds[0, 0])

        logger.info(f"Identifying approximate conditions with the highest expected hypervolume improvement")
        with np.errstate(divide="warn"):
            imp = Y_new_points_norm - np.amax(Y_norm) - self.xi
            Z = imp / Y_new_points_stddev_norm
            ei = imp * norm.cdf(Z) + Y_new_points_stddev_norm * norm.pdf(Z)
            ei[Y_new_points_stddev_norm == 0.0] = 0.0

        sorting_indices = ei[:, -1].argsort()
        index_array = np.arange(X_new_points.shape[0]).reshape(-1, 1)
        concat = np.hstack((X_new_points, Y_new_points, ei))
        concat_sorted = concat[sorting_indices][::-1]
        index_array_sorted = index_array[sorting_indices][::-1]

        # X_ei = concat_sorted[: self.num_candidates, :-2]
        # Y_ei = concat_sorted[: self.num_candidates, -2].flatten()

        X_ei = concat_sorted[:, :-2]
        Y_ei = concat_sorted[:, -2].flatten()

        X_ei_norm = ((X_ei - np.array(variables.lower_bounds))) / (
            np.array(variables.upper_bounds) - np.array(variables.lower_bounds)
        )

        # 0 index for the highest uncertainty one at the top of the array
        chosen_indices = [0]
        # cosine similarity index of 1.0 for the highest uncertainty one at the top of the array
        cosine_simularity_list = [1.0]
        # Comparing the X_values for the highest uncertainty with every other entry down the sorted X_array
        for index, x in enumerate(X_ei_norm):
            cosine_simularity_index = 1 - spatial.distance.cosine(X_ei_norm[0], x)
            # Only passes if the cosine similarity index of a given x row is greater than 0.1 different from all of the x's in the list
            if all((((x - cosine_simularity_index) ** 2) ** 0.5) > 0.1 for x in cosine_simularity_list):
                cosine_simularity_list.append(cosine_simularity_index)
                chosen_indices.append(index)

            if len(chosen_indices) == self.num_candidates:
                break

        X_ei = X_ei[chosen_indices]
        Y_ei = Y_ei[chosen_indices].astype(np.float64)

        # Writes the indexes of the selected candidates in the sample pool
        if isinstance(self.sampler, PoolBased):
            self.sampler.index = index_array_sorted[: self.num_candidates].flatten().tolist()

            logger.info("Conditions were successfully selected using expected improvement")
            return X_ei, Y_ei

        elif isinstance(self.sampler, SampleGenerator):
            # The objective function to be minimized for the scipy optimize minimize method
            def fun(X_norm):
                X_norm_reshape = X_norm.astype(np.float64).reshape(X_ei.shape)
                X = (X_norm_reshape * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))) + np.array(
                    variables.lower_bounds
                )
                if variables.num_cat_descriptor_var > 0:
                    X = variables.descriptor_to_name(X)

                Y, Y_stddev = objectives.evaluate(X)
                Y_norm = (Y - objectives.bounds[0, 0]) / (objectives.bounds[0, 1] - objectives.bounds[0, 0])
                # I think this should be fine to minmax normalise like this
                Y_stddev_norm = Y_stddev / (objectives.bounds[0, 1] - objectives.bounds[0, 0])

                with np.errstate(divide="warn"):
                    imp = Y_norm - np.amax(Y_norm) - self.xi
                    Z = imp / Y_stddev_norm
                    ei = imp * norm.cdf(Z) + Y_stddev_norm * norm.pdf(Z)
                    ei[Y_stddev_norm == 0.0] = 0.0

                # ei_mean = np.mean(ei)
                ei_sum = np.sum(ei)

                # return ei_mean
                return ei_sum

            # Convert the names of categorical variables to their descriptor values
            if variables.num_cat_descriptor_var > 0:
                X_ei = variables.categorical_transform(X_ei)

            # Normalise all variable values against the bounds. This will need to change if one-hot is implemented fully
            X_ei_norm = (X_ei - np.array(variables.lower_bounds)) / (
                np.array(variables.upper_bounds) - np.array(variables.lower_bounds)
            )

            logger.info(f"Optimising the variable location that provides the highest expected improvement")
            if constraints is None:
                method = "L-BFGS-B"
                consts = ()
                options = {"iprint": 99}
            else:
                method = "SLSQP"
                consts = constraints.create_scipy_constraints(self.num_candidates)
                options = {"disp": True}

            res = scipy_minimize(
                fun,
                X_ei_norm.flatten(),
                method=method,
                jac="3-point",
                bounds=Bounds(
                    lb=np.zeros_like(X_ei_norm).flatten(), ub=np.ones_like(X_ei_norm).flatten(), keep_feasible=True,
                ),
                constraints=consts,
                options=options,
            )

            if res.success == True:
                logger.info(
                    f"Successfully optimised the variable location that provides the highest expected improvement"
                )
                selected_X = res.x.astype(np.float64).reshape(X_ei.shape)
                selected_X = (
                    selected_X * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
                ) + np.array(variables.lower_bounds)

                if variables.num_cat_var > 0:
                    selected_X = variables.categorical_values_euc(selected_X)
                if variables.num_cat_descriptor_var > 0:
                    selected_X = variables.descriptor_to_name(selected_X)

                selected_Y, Y_new_stddev = objectives.evaluate(selected_X)

            else:
                logger.warning(
                    f"Optimisation of the variable location that provides the highest expected improvement failed. Will continue with the estimated position"
                )
                selected_X = X_ei
                selected_Y = Y_ei

            logger.info("Conditions were successfully selected using expected improvement")
            return selected_X, selected_Y

    def ehvi(
        self,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        constraints: Optional[ConstraintsList] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that determines the candidates with the best expected hypervolume improvement

        Parameters
        ----------
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest expected hypervolume improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest expected hypervolume improvement

        """
        logger.info(f"Identifying conditions with the highest expected hypervolume improvement")

        # Construct hypervolume reference point
        ref_point = self.build_ref_point(objectives.max_bool_dict)

        # logger.info(f"Generating {num_new_points} mixed integer Latin hypercube sampling points in the variable space")
        if isinstance(self.sampler, PoolBased):
            logger.info(f"Returning X values from the provided sample pool")
            X_new_points = self.sampler.X_pool
            index_combos = list(combinations(np.arange(X_new_points.shape[0]), self.num_candidates))
        elif isinstance(self.sampler, SampleGenerator):
            X_new_points = self.sampler.generate_samples(variables, constraints)

        Y_new_points, Y_new_points_stddev = objectives.evaluate(X_new_points)

        sign_adjusted_Y = self.Y_norm_minmax_transform(Y, objectives.bounds, objectives.max_bool_dict)
        sign_adjusted_Y_new_points = self.Y_norm_minmax_transform(
            Y_new_points, objectives.bounds, objectives.max_bool_dict
        )

        # Min_max normalisation of Y standard deviation
        # I think this should be fine to minmax normalise like this
        for obj_index, _ in enumerate(objectives.bounds):
            Y_new_points_stddev[:, obj_index] = Y_new_points_stddev[:, obj_index] / (
                objectives.bounds[obj_index, 1] - objectives.bounds[obj_index, 0]
            )

        # Creates a list of expected hypervolume improvements for all combinations of the new Y points
        logger.info(f"Identifying approximate conditions with the highest expected hypervolume improvement")
        ehvi = ExpectedHypervolumeImprovement(
            ref_point=torch.tensor(ref_point, dtype=torch.double), Y=torch.tensor(sign_adjusted_Y, dtype=torch.double),
        )
        ehvi_list = []
        X_new_points_combos = list(combinations(X_new_points, self.num_candidates))
        Y_new_points_combos = list(combinations(Y_new_points, self.num_candidates))
        Y_new_points_stddev_combos = list(combinations(Y_new_points_stddev, self.num_candidates))
        sign_adjusted_Y_new_points_combos = list(combinations(sign_adjusted_Y_new_points, self.num_candidates))
        for candidates_Y, candidates_Y_stddev in zip(sign_adjusted_Y_new_points_combos, Y_new_points_stddev_combos):
            try:
                val = ehvi.ehvi_calc(
                    Y_new=torch.tensor(np.array(candidates_Y), dtype=torch.double),
                    Y_new_stddev=torch.tensor(np.array(candidates_Y_stddev), dtype=torch.double),
                )
                ehvi_list.append(val)
            except ValueError:
                ehvi_list.append(torch.tensor(1e-8, dtype=torch.double))

        # Selects the variables and objective values of the candidate(s) that produced the highest expected hypervolume improvement
        max_idx = ehvi_list.index(max(ehvi_list))
        X_ehvi = np.array(X_new_points_combos[max_idx])
        Y_ehvi = np.array(Y_new_points_combos[max_idx])

        # Writes the indexes of the selected candidates in the sample pool
        if isinstance(self.sampler, PoolBased):
            self.sampler.index = list(index_combos[max_idx])

            logger.info("Conditions were successfully selected using expected hypervolume improvement")
            return X_ehvi, Y_ehvi

        elif isinstance(self.sampler, SampleGenerator):
            # The objective function to be minimized for the scipy optimize minimize method
            def fun(X_norm):
                X_norm_reshape = X_norm.astype(np.float64).reshape(X_ehvi.shape)
                X = (X_norm_reshape * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))) + np.array(
                    variables.lower_bounds
                )

                if variables.num_cat_descriptor_var > 0:
                    X = variables.descriptor_to_name(X)

                Y, Y_stddev = objectives.evaluate(X)

                sign_adjusted_Y = self.Y_norm_minmax_transform(Y, objectives.bounds, objectives.max_bool_dict)
                # Min_max normalisation of Y standard deviation
                # I think this should be fine to minmax normalise like this
                for obj_index, _ in enumerate(objectives.bounds):
                    Y_stddev[:, obj_index] = Y_stddev[:, obj_index] / (
                        objectives.bounds[obj_index, 1] - objectives.bounds[obj_index, 0]
                    )

                try:
                    fun_val = -1 * ehvi.ehvi_calc(
                        Y_new=torch.tensor(sign_adjusted_Y, dtype=torch.double),
                        Y_new_stddev=torch.tensor(Y_stddev, dtype=torch.double),
                    )
                except ValueError:
                    fun_val = -1 * torch.tensor(1e-8, dtype=torch.double)

                return fun_val.item()

            # Convert the names of categorical variables with their descriptor values
            if variables.num_cat_descriptor_var > 0:
                X_ehvi = variables.categorical_transform(X_ehvi)

            # Normalise all variable values against the bounds. This will need to change if one-hot is implemented fully
            X_ehvi_norm = (X_ehvi - np.array(variables.lower_bounds)) / (
                np.array(variables.upper_bounds) - np.array(variables.lower_bounds)
            )

            # Performs the scipy optimize minimize method
            logger.info(f"Optimising the variable location that provides the highest expected hypervolume improvement")
            if constraints is None:
                method = "L-BFGS-B"
                consts = ()
                options = {"iprint": 99}
            else:
                method = "SLSQP"
                consts = constraints.create_scipy_constraints(self.num_candidates)
                options = {"disp": True}

            res = scipy_minimize(
                fun,
                X_ehvi_norm,
                method=method,
                jac="3-point",
                bounds=Bounds(
                    lb=np.zeros_like(X_ehvi_norm).flatten(), ub=np.ones_like(X_ehvi_norm).flatten(), keep_feasible=True,
                ),
                constraints=consts,
                options=options,
            )

            if res.fun * -1 > max(ehvi_list).item():
                if res.success == True:
                    logger.info(
                        f"Successfully optimised the variable location that provides the highest expected hypervolume improvement"
                    )
                else:
                    logger.warning(
                        f"Attempts to optimise the variable location that provides the highest expected hypervolume improvement failed to converge but still improved and so will use the new X values"
                    )

                # Transform the optimised X value and calculate the corresponding objective Y values
                selected_X = res.x.astype(np.float64).reshape(X_ehvi.shape)
                selected_X = (
                    selected_X * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
                ) + np.array(variables.lower_bounds)

            else:
                logger.warning(
                    f"Attempts to optimise the variable location failed to improve the highest expected hypervolume improvement. Will continue with the estimated X values"
                )

                return X_ehvi, Y_ehvi

            if variables.num_cat_var > 0:
                selected_X = variables.categorical_values_euc(selected_X)
            if variables.num_cat_descriptor_var > 0:
                selected_X = variables.descriptor_to_name(selected_X)

            selected_Y, selected_Y_stddev = objectives.evaluate(selected_X)
            logger.info("Conditions were successfully selected using expected hypervolume improvement")

            return selected_X, selected_Y



# https://github.com/CDonnerer/xgboost-distribution
from __future__ import annotations

import os
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple

import numpy as np
import sklearn
from hyperopt import STATUS_OK, Trials, fmin, hp, tpe
from hyperopt.pyll.base import scope
from sklearn.model_selection import train_test_split
from xgboost_distribution import XGBDistribution

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.perf_metrics as pm
from nemo_bo.models.base.base_model import Base_Model
from nemo_bo.utils.data_proc import sort_train_test_split_shuffle

if TYPE_CHECKING:
    from nemo_bo.opt.objectives import RegressionObjective
    from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


class XGBoostModel(Base_Model):
    """

    Used to create a XGBoost Distribution model for the RegressionObjective

    See docstring for Base_Model __init__ function for information

    """

    def __init__(
        self, variables: VariablesList, objective: RegressionObjective, always_hyperparam_opt: bool = True, **kwargs
    ):
        super().__init__(variables, objective, always_hyperparam_opt)
        self.default_X_transform_type = "none"
        self.default_Y_transform_type = "none"
        self.include_validation = True
        self.name = "xgb"

    def fit(
        self, X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, sort_before_split: bool = True, **params
    ) -> None:
        """

        Function that is called to start the XGBoost Distribution model fitting procedure

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split for the validation set where applicable
        sort_before_split: bool, Default = True,
            Whether the inputted X and Y arrays should be sorted such that the Y-values are in ascending order before
            splitting into the training and test sets (and validation sets if applicable)
        params: Dict[str, Any]
            Dictionary that contains the hyperparameters for the model

        """
        if sort_before_split:
            (X_train, X_val, Y_train, Y_val,) = sort_train_test_split_shuffle(X, Y, test_ratio=test_ratio, seed=1)
        else:
            X_train, X_val, Y_train, Y_val = train_test_split(X, Y, test_size=test_ratio, random_state=1)

        Y_train = Y_train.astype("float")
        Y_val = Y_val.astype("float")

        self.X_train, self.X_val, self.Y_train, self.Y_val = (
            X_train,
            X_val,
            Y_train,
            Y_val,
        )

        if self.variables.num_cat_var > 0:
            self.X_train = self.variables.categorical_transform(X_train).astype("float")
            self.X_val = self.variables.categorical_transform(X_val).astype("float")
        self.X_train, self.Y_train = self.transform_by_predictor_type(self.X_train, Y=self.Y_train)
        self.X_val, self.Y_val = self.transform_only_by_predictor_type(self.X_val, self.Y_val)

        for key in params:
            if key == "max_depth":
                params["max_depth"] = int(params["max_depth"])
            if key == "n_estimators":
                params["n_estimators"] = int(params["n_estimators"])

        self.model = XGBDistribution(early_stopping_rounds=10, **params)

        self.fit_model(params)

        self.Y_train_pred, self.Y_train_pred_stddev = self.predict(X_train)
        self.Y_val_pred, self.Y_val_pred_stddev = self.predict(X_val)
        self.performance_metrics = pm.all_performance_metrics_train_val(
            Y_train, self.Y_train_pred, Y_val, self.Y_val_pred
        )

        self.Y_train_error = 1.96 * self.Y_train_pred_stddev
        self.Y_val_error = 1.96 * self.Y_val_pred_stddev

    def fit_model(self, params: Dict[str, Any]) -> None:
        """

        Called by the fit and cv-related functions for fitting the XGBoost Distribution model using pre-processed X
        and Y training data in the class instance

        """
        for key in params:
            if key == "max_depth":
                params["max_depth"] = int(params["max_depth"])
            if key == "n_estimators":
                params["n_estimators"] = int(params["n_estimators"])

        self.model = XGBDistribution(early_stopping_rounds=10, **params)

        # Re-attempts fitting
        for x in range(10):
            try:
                self.model.fit(
                    self.X_train, self.Y_train, eval_set=[(self.X_val, self.Y_val)], verbose=False,
                )
                break
            except np.linalg.LinAlgError as e:
                print(e)
            except TypeError as e:
                print(e)
            except sklearn.exceptions.NotFittedError as e:
                print(e)

    def predict(self, X: np.ndarray, X_transform: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """

        Determines the model output and its standard deviation of the X array passed into the function

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation
        X_transform: bool, Default = True
            For the default models provided natively in NEMO, the initial transformation of categorical variables and
            transformations related to standardisation/normalisation can be switched off by passing X_transform = False

        Returns
        ----------
        Y_mean: np.ndarray
            The predictive mean of the objective calculated using the model and the inputted X array. The outputted
            values have been un-transformed
        Y_pred_stddev: np.ndarray
            The predictive standard deviation of the objective calculated using the model and the inputted X array. The
            outputted values have been un-transformed

        """
        # X needs to be a 2D array
        if X.ndim == 1:
            X = X.reshape(1, -1)

        if X_transform:
            if self.variables.num_cat_var > 0:
                X = self.variables.categorical_transform(X).astype("float")
            X = self.transform_only_by_predictor_type(X)

        Y_pred = self.model.predict(X.astype(float))
        Y_mean = Y_pred.loc
        Y_pred_upper = Y_mean + (1.96 * Y_pred.scale)

        Y_pred = self.objective.inverse_transform(Y_mean)
        Y_pred_upper = self.objective.inverse_transform(Y_pred_upper)
        Y_pred_stddev = (np.subtract(Y_pred_upper, Y_pred)) / 1.96

        return Y_mean.flatten(), Y_pred_stddev.flatten()

    @staticmethod
    def default_params():
        """

        Function that returns the default hyperparameters to use for the hyperparameter optimisation of XGBoost
        Distribution models using the hyperopt package. Returns a dictionary that is used for the space argument in
        the hyperopt.fmin function

        """
        return {
            "eta": hp.uniform("eta", 0.01, 0.5),
            "max_depth": scope.int(hp.quniform("max_depth", 3, 18, 1)),
            "min_child_weight": hp.uniform("min_child_weight", 0.01, 0.75),
            "colsample_bytree": hp.uniform("colsample_bytree", 0.5, 1),
        }

    def hyperparam_opt(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float,
        sort_before_split: bool = True,
        predictor_params_dict: Optional[Dict[str, Dict[str, Callable]]] = None,
        **kwargs,
    ) -> Tuple[XGBoostModel, Dict[str, Any]]:
        """

        Function to be called for performing hyperparameter optimisation for XGBoost Distribution models using the
        hyperopt package. Returns the model fitted with the best hyperparameters and a dictionary containing the best
        hyperparameters

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float
            The proportion of inputted X and Y arrays to be split for the validation set where applicable
        sort_before_split: bool, Default = True,
            Whether the inputted X and Y arrays should be sorted such that the Y-values are in ascending order before
            splitting into the training and test sets (and validation sets if applicable)
        predictor_params_dict: Dict[str, Dict[str, Callable]], Default = None
            Dictionary with a key 'xgb' and a corresponding value that is a dictionary that is used for the space
            argument in the hyperopt.fmin function

        Returns
        -------
        model: XGBoostModel
            The model fitted with the best hyperparameters found using the hyperopt package

        model_params: Dict[str, Any]
            Dictionary containing the best hyperparameters found using the hyperopt package

        """
        if predictor_params_dict is None:
            predictor_params_dict = self.default_params()
        else:
            predictor_params_dict = predictor_params_dict.get(self.name, self.default_params())

        # if self.objective.hyperopt_evals is None:
        #     hyperopt_evals = 1
        # else:
        #     hyperopt_evals = self.objective.hyperopt_evals.get(self.name, 1)
        if self.objective.hyperopt_evals is None:
            hyperopt_evals = 100
        else:
            hyperopt_evals = self.objective.hyperopt_evals.get(f"{self.name}", 100)

        def func(X: np.ndarray, Y: np.ndarray, model_params: Dict[str, Any]) -> Dict[str, Any]:
            cv_results = self.cv(
                X, Y, model_params, test_ratio=test_ratio, sort_before_split=sort_before_split, **kwargs
            )

            return {
                "loss": cv_results["Mean Test RMSE"],
                "status": STATUS_OK,
            }

        trials = Trials()
        model_params = fmin(
            fn=partial(func, X, Y),
            space=predictor_params_dict,
            algo=tpe.suggest,
            max_evals=hyperopt_evals,
            trials=trials,
        )

        model = self.new_instance(self.__class__, **kwargs)
        model.fit(X, Y, test_ratio=test_ratio, **model_params, **kwargs)

        logger.info(f"Completed hyperparameter opt")
        return model, model_params



import os
from attrs import define, field
from typing import Callable, List, Tuple, Union

import numpy as np
import torch
from torch import Tensor

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@define
class LinearConstraint:
    """

    Implements a linear constraint on the input variables for the optimisation. For example, linear constraints could be:

        a(F1) + b(F2) = 0
        a(F1) + b(F2) >= 0

    where a and b are coefficients for the features F1 and F2 respectively, and the right-hand side of the equation is
    0. Using an equals sign, =, would form a linear equality constraint whereas defining as 'greater than or equal to'
    forms an inequality constraint.

    By default in NEMO, all inequality constraints must be defined as 'greater than or equal to'. Note: This aligns
    with how constraints are defined in the scipy package, but is the opposite when compared to the pymoo package

    Parameters
    ----------
    constraint_type: str
        Defines whether the linear constraint should be an equality or inequality linear constraint by assigning as
        either 'eq' or 'ineq' respectively
    variables: VariablesList
        VariablesList object that contains information about all variables
    features: List[str]
        The name of the features to constrain, e.g. ['F1', 'F2'] for the example equation shown above
    coefficients: List[Union[int, float]]
        The coefficient for every features in the constraint. The values in the provided list must be given in
        the same order as the provided features to constrain, e.g. [a, b] for the example equation shown above
    rhs: Union[int, float]
        The value to set for the right-hand side of the equation, e.g. 0 for the example equation shown above

    """

    constraint_type: str
    variables: VariablesList
    features: List[str]
    coefficients: List[Union[int, float]]
    rhs: Union[int, float]
    index: List[int] = field(init=False)

    def __attrs_post_init__(self):
        self.index = [self.variables.names.index(f) for f in self.features]

    def build_polytope_sampler_constraint(self) -> [Tuple[Tensor, Tensor, float]]:
        """

        Function that creates the constraint tuple that is used in the polytope sampler

        """
        # How to handle this index when the continuous variable is after a categorical with descriptors?! Need to count the number of descriptors?
        # Can we somehow use the headings + descriptors attribute that we used for the continuous permutation feature importance?!

        return (torch.tensor(self.index), torch.tensor(self.coefficients), self.rhs)

    def build_scipy_constraint(self, add_to_index: int, num_candidates: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array
        num_candidates: int
            The number of candidates desired per iteration from the optimisation

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        new_index = [i + add_to_index for i in self.index]

        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # Saves the original shape of the X array
            x_shape = x.shape

            # Reshapes the X array
            x = x.reshape(num_candidates, -1)

            # Un-normalise the variables
            x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
                self.variables.lower_bounds
            )

            # Re-shape the X array back to the original shape
            x = x.reshape(x_shape)

            # Evaluate the X array for the constraint
            return np.sum(x[new_index] * self.coefficients) - self.rhs

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # Un-normalise the variables
        x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Multiply coefficients and rhs by -1 because pymoo uses the less than or equal to inequality operator for its constraints
        if self.constraint_type == "ineq":
            return np.sum(x[:, self.index] * [-coeff for coeff in self.coefficients], axis=1) - (self.rhs * -1)
        else:
            # an equality constraint can be expressed in this manner by squaring the inequality expression
            # See https://pymoo.org/misc/constraints.html for more details
            return (np.sum(x[:, self.index] * [-coeff for coeff in self.coefficients], axis=1) - (self.rhs * -1)) ** 2

    def bool_evaluate_constraint(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether the equality or inequality linear constraint is met and returns a boolean

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        if self.constraint_type == "ineq":
            return np.around(np.sum(X[:, self.index] * self.coefficients, axis=1), 4) >= round(float(self.rhs), 4)
        else:
            return np.around(np.sum(X[:, self.index] * self.coefficients, axis=1), 4) == round(float(self.rhs), 4)


@define
class NonLinearConstraint:
    """

    Implements a linear constraint on the input variables for the optimisation. For example, linear constraints could be:

        a(F1)^x + b(F2)^y = 0
        a(F1)^x + b(F2)^y >= 0

    where a and b are coefficients, and x and y are powers (exponents) for the features F1 and F2 respectively, and the
    right-hand side of the equation is 0. Using an equals sign, =, would form a linear equality constraint whereas
    defining as 'greater than or equal to' forms an inequality constraint.

    By default in NEMO, all inequality constraints must be defined as 'greater than or equal to'. Note: This aligns
    with how constraints are defined in the scipy package, but is the opposite when compared to the pymoo package

    Parameters
    ----------
    constraint_type: str
        Defines whether the linear constraint should be an equality or inequality linear constraint by assigning as
        either 'eq' or 'ineq' respectively
    variables: VariablesList
        VariablesList object that contains information about all variables
    features: List[str]
        The name of the features to constrain, e.g. ['F1', 'F2'] for the example equation shown above
    coefficients: List[Union[int, float]]
        The coefficient for every features in the constraint. The values in the provided list must be given in
        the same order as the provided features to constrain, e.g. [a, b] for the example equation shown above
    powers: List[int]
        The power (exponent) for every features in the constraint. The values in the provided list must be given in
        the same order as the provided features to constrain, e.g. [x, y] for the example equation shown above
    rhs: Union[int, float]
        The value to set for the right-hand side of the equation, e.g. 0 for the example equation shown above

    """

    constraint_type: str
    variables: VariablesList
    features: List[str]
    coefficients: List[Union[int, float]]
    powers: List[int]
    rhs: Union[int, float]
    index: List[int] = field(init=False)

    def __attrs_post_init__(self):
        self.index = [self.variables.names.index(f) for f in self.features]

    def build_polytope_sampler_constraint(self):
        raise RuntimeError("The polytope sampler cannot be used with non-linear constraints")

    def build_scipy_constraint(self, add_to_index: int, num_candidates: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array
        num_candidates: int
            The number of candidates desired per iteration from the optimisation

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        new_index = [i + add_to_index for i in self.index]

        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # Saves the original shape of the X array
            x_shape = x.shape

            # Reshapes the X array
            x = x.reshape(num_candidates, -1)

            # Un-normalise the variables
            x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
                self.variables.lower_bounds
            )

            # Re-shape the X array back to the original shape
            x = x.reshape(x_shape)

            # Evaluate the X array for the constraint
            lhs_eval_list = []
            for i, c, p in zip(new_index, self.coefficients, self.powers):
                lhs_eval_list.append((((x[i]) ** p) * c))

            return sum(lhs_eval_list) - self.rhs

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # Un-normalise the variables
        x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Multiply coefficients and rhs by -1 because pymoo uses the less than or equal to inequality operator for its constraints
        lhs_eval_list = []
        if self.constraint_type == "ineq":
            for i, c, p in zip(self.index, self.coefficients, self.powers):
                lhs_eval_list.append((((x[:, i]) ** p) * -c))

            return np.sum(lhs_eval_list, axis=0) - (self.rhs * -1)

        else:
            # an equality constraint can be expressed in this manner by squaring the inequality expression
            # See https://pymoo.org/misc/constraints.html for more details
            for i, c, p in zip(self.index, self.coefficients, self.powers):
                lhs_eval_list.append((((x[:, i]) ** p) * -c))

            return (np.sum(lhs_eval_list, axis=0) - (self.rhs * -1)) ** 2

    def bool_evaluate_constraint(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether the equality or inequality linear constraint is met and returns a boolean

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        if self.constraint_type == "ineq":
            return np.around(np.sum(X[:, self.index] * self.coefficients, axis=1), 4) >= round(float(self.rhs), 4)
        else:
            return np.around(np.sum(X[:, self.index] * self.coefficients, axis=1), 4) == round(float(self.rhs), 4)


@define
class MaxActiveFeaturesConstraint:
    """

    Implements a constraint that limits the number of variables/features that have values greater than 0. For example,
    if max_active = 2, the examples 1 and 2 below would pass the constraint whereas example 3 would fail the
    constraint:

        - Example with 4 variables in the X array
        - Only need to consider variables at index 0, 1, 2
        - Ensure that only 2 variables from indexes 0, 1, and 2 are above 0 (max_active = 2)

        1. X = [0.5, 0.7, 0.0, 1.0] -> 2 variables from indexes 0, 1, 2 are above 0 -> Pass
        2. X = [0.5, 0.0, 0.3, 1.0] -> 2 variables from indexes 0, 1, 2 are above 0 -> Pass
        3. X = [0.5, 0.7, 0.3, 1.0] -> 3 variables from indexes 0, 1, 2 are above 0 -> Fail

    Parameters
    ----------
    variables: VariablesList
        VariablesList object that contains information about all variables
    features: List[str]
        The name of the features to consider in the constraint, e.g. ['F1', 'F2']
    max_active: int
        The maximum number of features from the provided names that can be above 0

    """

    variables: VariablesList
    features: List[str]
    max_active: int
    index: List[int] = field(init=False)

    def __attrs_post_init__(self):
        self.index = [self.variables.names.index(f) for f in self.features]

    def build_polytope_sampler_constraint(self):
        raise RuntimeError("The polytope sampler cannot be used with max active number of features constraints")

    def build_scipy_constraint(self, add_to_index: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        new_index = [i + add_to_index for i in self.index]

        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # The values in the x array that come from the scipy minimize function need to be normalised between 0 and 1
            # Returns a one where the values are above 0 and zero where they equal 0
            x = np.where(x > 0, 1, 0)

            # Evaluate the one-zero X array for the constraint
            return np.sum(x[new_index] * ([-1] * len(new_index))) - (self.max_active * -1)

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # The values in the x array that come from pymoo need to be normalised between 0 and 1
        # Multiply coefficients and rhs by -1 because pymoo uses the less than or equal to inequality operator for its constraints
        X = np.where(x > 0, 1, 0)
        return np.sum(X[:, self.index], axis=1) - self.max_active

    def bool_evaluate_constraint(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether the constraint is met and returns a boolean array

        Parameters
        ----------
        X: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        X = np.where(X > 0, 1, 0)
        return np.sum(X[:, self.index], axis=1) <= self.max_active


@define
class CategoricalConstraint:
    """

    Implements a constraint that prevents chosen categorical levels (e.g. chemicals) from two variables/features from
    being simultaneously selected. For example, the examples 1 and 2 below would pass the constraint whereas examples
    3 and 4 would fail the constraint:

        - Exampler where variables at indexes 0 and 1 are 'Solvent' and 'Base' respectively
        - In the 'Solvent' variable, the categorical level called 'Water' cannot be selected at the same time as
        categorical levels called 'Sodium' or 'Potassium' in the 'Base' variable

        1. X = ['Water', 'Triethylamine'] -> Pass
        2. X = ['Toluene', 'Sodium'] -> Pass
        3. X = ['Water', 'Sodium'] -> Fail
        4. X = ['Water', 'Potassium'] -> Fail

    Parameters
    ----------
    variables: VariablesList
        VariablesList object that contains information about all variables
    feature1: str
        The name of one of the feature to consider in the constraint, e.g. 'Solvent'
    feature2: str
        The name of the other feature to consider in the constraint, e.g. 'Base'
    categorical_levels1: List[Union[str, int, float]]
        The categorical levels in feature1 that cannot be simultaneously selected with categorical levels in feature2,
        e.g. ['Water']
    categorical_levels2: List[Union[str, int, float]]
        The categorical levels in feature2 that cannot be simultaneously selected with categorical levels in feature1,
        e.g. ['Sodium', 'Potassium']

    """

    variables: VariablesList
    feature1: str
    feature2: str
    categorical_levels1: List[Union[str, int, float]]
    categorical_levels2: List[Union[str, int, float]]
    index: List[int] = field(init=False)

    def __attrs_post_init__(self):
        self.index1 = [self.variables.names.index(self.feature1)]
        self.index2 = [self.variables.names.index(self.feature2)]

    def build_polytope_sampler_constraint(self):
        raise RuntimeError("The polytope sampler cannot be used with categorical constraints")

    def build_scipy_constraint(self, add_to_index: int, num_candidates: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """

        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # Saves the original shape of the X array
            x_shape = x.shape

            # Reshapes the X array
            x = x.reshape(num_candidates, -1)

            # Un-normalise the variables
            x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
                self.variables.lower_bounds
            )

            # Converts the descriptor values generated by the scipy minimize function into the categorical level that
            # is the closest by Euclidean distance
            x = self.variables.descriptor_to_name(x)

            # Checks if the forbidden categorical levels are in the chosen features
            for row_index, (x1, x2) in enumerate(
                zip(x[:, self.index1 + add_to_index], x[:, self.index2 + add_to_index])
            ):
                if x1 in self.categorical_levels1:
                    x[row_index, self.index1 + add_to_index] = 1
                if x2 in self.categorical_levels2:
                    x[row_index, self.index2 + add_to_index] = 1

            # Re-shape the X array back to the original shape
            x = x.reshape(x_shape)

            # Evaluate the X array for the constraint
            return (x[{self.index1 + add_to_index}] * -1) + (x[{self.index2 + add_to_index}] * -1) - (-1)

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # Un-normalise the variables
        x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Converts the descriptor values generated by the scipy minimize function into the categorical level that
        # is the closest by Euclidean distance
        x = self.variables.descriptor_to_name(x)

        # Checks if the forbidden categorical levels are in the chosen features
        for row_index, (x1, x2) in enumerate(zip(x[:, self.index1], x[:, self.index2])):
            if x1 in self.categorical_levels1:
                x[row_index, self.index1] = 1
            if x2 in self.categorical_levels2:
                x[row_index, self.index2] = 1

        return np.sum(x[:, [self.index1, self.index2]], axis=1) - 1

    def bool_evaluate_constraint(self, X: np.ndarray, descriptor_to_name: bool = True) -> np.ndarray:
        """

        Evaluate whether the constraint is met and returns a boolean array

        Parameters
        ----------
        X: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values
        descriptor_to_name: bool
            If descriptors are used to describe the categorical variables in the X array, this needs to be set to True

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        if descriptor_to_name:
            # Converts the descriptor values generated by the scipy minimize function into the categorical level that
            # is the closest by Euclidean distance
            X = self.variables.descriptor_to_name(X)

        # Checks if the forbidden categorical levels are in the chosen features
        for row_index, (x1, x2) in enumerate(zip(X[:, self.index1], X[:, self.index2])):
            if x1 in self.categorical_levels1:
                X[row_index, self.index1] = 1
            if x2 in self.categorical_levels2:
                X[row_index, self.index2] = 1

        return np.sum((X[:, [self.index1, self.index2]]), axis=1) <= 1


@define
class ConstraintsList:
    """

    Class that contains information and functions for all constraints

    Parameters
    ----------
    constraints: List[Union[LinearConstraint, NonLinearConstraint, MaxActiveFeaturesConstraint, CategoricalConstraint]]
        List of all constraints

    """

    constraints: List[Union[LinearConstraint, NonLinearConstraint, MaxActiveFeaturesConstraint, CategoricalConstraint]]
    n_constr: int = field(init=False)
    variables: VariablesList = field(init=False)

    def __attrs_post_init__(self):
        self.n_constr = len(self.constraints)
        self.variables = self.constraints[0].variables

    def create_scipy_constraints(self, num_candidates: int) -> List[Tuple[Tensor, Tensor, float]]:
        """

        Creates the constraint functions that are supplied into the scipy minimize function for all constraints

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array

        Returns
        -------
        scipy_constraints: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        scipy_constraints = []
        for constr in self.constraints:
            for candidate_number in range(num_candidates):
                if not isinstance(constr, CategoricalConstraint):
                    add_to_index = candidate_number * constr.variables.n_var
                else:
                    add_to_index = candidate_number * len(constr.variables.variables)
                scipy_constraints.append(
                    {
                        "type": constr.constraint_type,
                        "fun": constr.build_scipy_constraint(add_to_index, num_candidates)
                        if not isinstance(constr, MaxActiveFeaturesConstraint)
                        else constr.build_scipy_constraint(add_to_index),
                    }
                )

        return scipy_constraints

    def bool_evaluate_constraints(self, X: np.ndarray, descriptor_to_name: bool = True) -> np.ndarray:
        pass_fail = np.zeros((X.shape[0], self.n_constr))

        # Needed for all constraints other than CategoricalConstraint because it ensures that the index is correct.
        # Index must refer to the index when categoricalwithdescriptors have descriptors present (not names)
        if self.variables.num_cat_var > 0:
            X = self.variables.categorical_transform(X)

        for constr_index, constr in enumerate(self.constraints):
            if not isinstance(constr, CategoricalConstraint):
                pass_fail[:, constr_index] = constr.bool_evaluate_constraint(X)
            else:
                pass_fail[:, constr_index] = constr.bool_evaluate_constraint(X, descriptor_to_name)

        return np.all(pass_fail, axis=1)

    def bool_evaluate_constraints_categorical(self, X: np.ndarray) -> np.ndarray:
        pass_fail = np.zeros((X.shape[0], self.n_constr))
        for constr_index, constr in enumerate(self.constraints):
            X = constr.variables.categorical_transform(X)
            pass_fail[:, constr_index] = constr.bool_evaluate_constraint(X)

        return np.all(pass_fail, axis=1)



import os
from attrs import define, field
from typing import Optional, Tuple

import numpy as np
import torch

from botorch.test_functions.synthetic import (
    Ackley,
    DixonPrice,
    Griewank,
    Levy,
    Michalewicz,
    Rastrigin,
    Rosenbrock,
    StyblinskiTang,
)

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.opt.objectives import CalculableObjective, ObjectivesList, RegressionObjective
from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@define
class Benchmark:
    """

    Benchmark functions are typically used to simulate the outcomes of experiments in a closed-loop manner and therefore
    can be helpful to evaluate the quality of an optimisation (inferred from the effectiveness of the utilised ML
    model(s) and/or acquisition function to identify the optimum)

    Parameters
    ----------
    variables: VariablesList
        VariablesList object that contains information about all variables
    objectives: ObjectivesList
        ObjectivesList object that contains information about all objectives

    """

    def evaluate(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        Returns the corresponding output values and standard deviations using the inputted X array and benchmark method

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation

        """
        pass


class SyntheticBenchmark:
    """

    Adopts the following synthetic functions from the BoTorch package to use as functions in a benchmark. The
    recommended bounds for each dimension is shown in parentheses:

    1. Ackley (-32.768, 32.768)
    2. DixonPrice (-10.0, 10.0)
    3. Griewank (-600.0, 600.0)
    4. Levy (-10.0, 10.0)
    5. Michalewicz (0.0, math.pi)
    6. Rastrigin (-5.12, 5.12)
    7. Rosenbrock (-5.0, 10.0)
    8. StyblinskiTang (-5.0, 5.0)

    Parameters
    ----------
    function_name: str
        The name of the synthetic function to use ('ackley', 'dixon-price', 'griewank', 'levy', 'michalewicz',
        'rastrigin', 'rosenbrock', or 'styblinski-tang')
    dim: int
        The number of input dimensions
    noise_std: float = None
        Standard deviation of the observation noise

    """

    def __init__(self, function_name: str, dim: int, noise_std: Optional[float] = None):
        if function_name == "ackley":
            self.synthetic_function = Ackley(dim=dim, noise_std=noise_std)
        if function_name == "dixon-price":
            self.synthetic_function = DixonPrice(dim=dim, noise_std=noise_std)
        if function_name == "griewank":
            self.synthetic_function = Griewank(dim=dim, noise_std=noise_std)
        if function_name == "levy":
            self.synthetic_function = Levy(dim=dim, noise_std=noise_std)
        if function_name == "michalewicz":
            self.synthetic_function = Michalewicz(dim=dim, noise_std=noise_std)
        if function_name == "rastrigin":
            self.synthetic_function = Rastrigin(dim=dim, noise_std=noise_std)
        if function_name == "rosenbrock":
            self.synthetic_function = Rosenbrock(dim=dim, noise_std=noise_std)
        if function_name == "styblinski-tang":
            self.synthetic_function = StyblinskiTang(dim=dim, noise_std=noise_std)

        self.fitted = True

    def evaluate(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        Calculate the corresponding output values using the desired synthetic function and the inputted X array

        Parameters
        ----------
        X: np.ndarray
            X array that contains only numerical data

        """
        Y = self.synthetic_function.evaluate_true(torch.tensor(X)).cpu().detach().numpy()

        return Y, np.zeros_like(Y)


@define
class ModelBenchmark(Benchmark):
    """

    This type of benchmark object is used to accurately predict the outcome of optimisation suggestions as they typically
    utilise all available data for regression objectives. This method will also work for the calculable objectives in
    the optimisation problem.

    Parameters
    ----------
    variables: VariablesList
        VariablesList object that contains information about all variables
    objectives: ObjectivesList
        ObjectivesList object that contains information about all objectives

    """

    variables: VariablesList
    objectives: ObjectivesList
    fitted: bool = field(init=False)

    def __attrs_post_init__(self):
        self.fitted = False

    def fit(self, X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, **kwargs) -> None:
        """

        Fits ML models for all specified regression objectives to be used a benchmark functions

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data for all calculable and regression objectives, i.e.
            The data has not yet undergone any transformations related to standardisation/normalisation
        test_ratio: float
            The proportion of inputted X and Y arrays to be split for the validation set where applicable

        """
        logger.info(
            "Performing hyperparameter and model search for the best predictor models for use as benchmarkers for the regression objectives"
        )

        # Fit the ML regression models
        self.objectives.fit(
            X, Y, self.variables, model_search_bool=True, test_ratio=test_ratio, **kwargs,
        )
        self.fitted = True

        logger.info(
            f"Identified the best predictor models and hyperparameters for the benchmarkers for the regression objectives"
        )

    def evaluate(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        Returns the corresponding output values and standard deviations using the inputted X array and benchmark models
        and calculable objectives

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation

        Returns
        ----------
        Y: np.ndarray
            Complete Y array of output values from all calculable and regression objectives using the provided
            X array
        Y_pred_stddev: np.ndarray
            Complete array of corresponding standard deviations for the predicted Y-values of all calculable and
            regression objectives using the provided X array

        """
        Y = np.zeros((X.shape[0], self.objectives.n_obj), dtype=np.float64)
        Y_stddev = np.zeros_like(Y)
        for obj_index, obj in enumerate(self.objectives.objectives):
            if isinstance(obj, RegressionObjective):
                Y[:, obj_index], Y_stddev[:, obj_index] = obj.predict(X)
            elif isinstance(obj, CalculableObjective):
                Y[:, obj_index], Y_stddev[:, obj_index] = obj.calculate(X)

        return Y.astype("float"), Y_stddev.astype("float")



import math
from typing import List, Union

import numpy as np
from numba import cuda
from sklearn.preprocessing import MinMaxScaler, StandardScaler


class Transformations:
    """

    Class for storing and applying data transformations

    """

    def standardscaler(self, data: np.ndarray) -> np.ndarray:
        """

        Standardises the data by utilising sklearn.preprocessing.StandardScaler to transform the data such that its
        distribution will have a mean value 0 and standard deviation of 1. For multivariate data, this is done
        feature-wise, i.e. independently for each column of the data

        Parameters
        ----------
        data: np.ndarray
            np.ndarray with shape (n_samples, n_features) containing the data used to compute the mean and standard
            deviation used for scaling along the features axis

        Returns
        -------
        data_standardised : np.ndarray
            np.ndarray with shape (n_samples, n_features) of standardised data

        """
        if data.ndim == 1:
            data = data.reshape(-1, 1)
        self.scaler = StandardScaler()
        data_standardised = self.scaler.fit_transform(data)

        return data_standardised

    def standardscaler_transform_only(self, data: np.ndarray) -> np.ndarray:
        """

        Standardises the data using the previously determined mean and standard deviation. For multivariate data,
        this is done feature-wise, i.e. independently for each column of the data

        Parameters
        ----------
        data: np.ndarray
            np.ndarray with shape (n_samples, n_features) containing the data to be transformed by standardisation
            along the features axis

        Returns
        -------
        data_standardised : np.ndarray
            np.ndarray with shape (n_samples, n_features) of standardised data

        """
        if data.ndim == 1:
            data = data.reshape(-1, 1)
        data_standardised = self.scaler.transform(data)

        return data_standardised

    def inverse_standardscaler(self, data_standardised: np.ndarray) -> np.ndarray:
        """

        Un-standardises the data using the previously determined mean and standard deviation. For multivariate data,
        this is done feature-wise, i.e. independently for each column of the data

        Parameters
        ----------
        data_standardised: np.ndarray
            np.ndarray with shape (n_samples, n_features) containing the data to be transformed by un-standardising
            along the features axis

        Returns
        -------
        data : np.ndarray
            np.ndarray with shape (n_samples, n_features) of un-standardised data

        """
        if data_standardised.ndim == 1:
            data_standardised = data_standardised.reshape(-1, 1)
        data = self.scaler.inverse_transform(data_standardised)

        return data

    def minmaxscaler(self, data: np.ndarray) -> np.ndarray:
        """

        Min-max normalises the data by utilising sklearn.preprocessing.MinMaxScaler to transform the data by scaling
        and translating each feature individually such that it is in the given range on the training set, e.g. between
        zero and one. For multivariate data, this is done feature-wise, i.e. independently for each column of the data

        Parameters
        ----------
        data : array-like with shape (n_samples, n_features) containing the data used to identify the minimum and
        maximum values to be used for min-max normalisation scaling along the features axis.

        Returns
        -------
        data_normalised : numpy.ndarray type of normalised data

        """

        if data.ndim == 1:
            data = data.reshape(-1, 1)
        self.scaler = MinMaxScaler()
        data_normalised = self.scaler.fit_transform(data)

        return data_normalised

    def minmaxscaler_transform_only(self, data: np.ndarray) -> np.ndarray:
        """

        Min-max normalises the data using the previously identified minimum and maximum values. For multivariate data,
        this is done feature-wise, i.e. independently for each column of the data

        Parameters
        ----------
        data: np.ndarray
            np.ndarray with shape (n_samples, n_features) containing the data to be transformed by min-max normalisation
            along the features axis

        Returns
        -------
        data_normalised : np.ndarray
            np.ndarray with shape (n_samples, n_features) of normalised data

        """
        if data.ndim == 1:
            data = data.reshape(-1, 1)
        data_normalised = self.scaler.transform(data)

        return data_normalised

    def inverse_minmaxscaler(self, data_normalised: np.ndarray) -> np.ndarray:
        """

        Un-normalises the data using the previously identified minimum and maximum values. For multivariate data,
        this is done feature-wise, i.e. independently for each column of the data

        Parameters
        ----------
        data_standardised: np.ndarray
            np.ndarray with shape (n_samples, n_features) containing the data to be transformed by un-normalising
            along the features axis

        Returns
        -------
        data : np.ndarray
            np.ndarray with shape (n_samples, n_features) of un-normalised data

        """
        if data_normalised.ndim == 1:
            data_normalised = data_normalised.reshape(-1, 1)
        data = self.scaler.inverse_transform(data_normalised)

        return data

    def numba_cuda_1d_kernel(self, array: np.ndarray) -> None:
        # Kernel declaration and invocation for a 1D array and grid of threads
        # The number of threads to assign to each block when instantiating the kernel
        self.threadsperblock1d = 1024
        # Defines number of thread blocks on the grid
        self.blockspergrid1d = (array.shape[0] + (self.threadsperblock1d - 1)) // self.threadsperblock1d

    def numba_cuda_2d_kernel(self, array: np.ndarray) -> None:
        # Kernel declaration and invocation for a 2D array and grid of threads
        # The number of threads to assign to each block when instantiating the kernel. Each value of 32 is equal to the square root of the maximum number of threads per block on the computer. In this case the square root of 1024
        self.threadsperblock2d = (
            32,
            32,
        )
        # Number of thread blocks in the x direction. FYI: The product of the number of thread blocks (or “blocks per grid”) and a number of threads per block will give the total number of threads launched.
        blockspergrid_x = math.ceil(array.shape[0] / self.threadsperblock2d[0])
        # Number of thread blocks in the y direction.
        blockspergrid_y = math.ceil(array.shape[1] / self.threadsperblock2d[1])
        # Defines number of thread blocks on the xy grid
        self.blockspergrid2d = (
            blockspergrid_x,
            blockspergrid_y,
        )

    def X_min_max_norm_cuda_run(
        self,
        X: np.ndarray,
        x_max: Union[List[Union[int, float]], np.ndarray],
        x_min: Union[List[Union[int, float]], np.ndarray],
    ) -> np.ndarray:
        X_norm = np.zeros_like(X)
        self.numba_cuda_2d_kernel(X)
        self.X_min_max_norm_cuda[self.blockspergrid2d, self.threadsperblock2d](X, X_norm, x_max, x_min)

        return X_norm

    @staticmethod
    @cuda.jit
    def X_min_max_norm_cuda(
        X: np.ndarray,
        X_norm: np.ndarray,
        x_max: Union[List[Union[int, float]], np.ndarray],
        x_min: Union[List[Union[int, float]], np.ndarray],
    ) -> None:
        # Normalise the features by applying min-max feature scaling
        # Increments through a 2D grid
        x, y = cuda.grid(2)
        if x < X.shape[0] and y < X.shape[1]:
            X_norm[x, y] = (X[x][y] - x_min[y]) / (x_max[y] - x_min[y])

    def X_inverse_min_max_norm_cuda_run(
        self,
        X_norm: np.ndarray,
        x_max: Union[List[Union[int, float]], np.ndarray],
        x_min: Union[List[Union[int, float]], np.ndarray],
    ) -> np.ndarray:
        X = np.zeros_like(X_norm)
        self.numba_cuda_2d_kernel(X_norm)
        self.X_inverse_min_max_norm_cuda[self.blockspergrid2d, self.threadsperblock2d](X_norm, X, x_max, x_min)

        return X

    @staticmethod
    @cuda.jit
    def X_inverse_min_max_norm_cuda(
        X_norm: np.ndarray,
        X: np.ndarray,
        x_max: Union[List[Union[int, float]], np.ndarray],
        x_min: Union[List[Union[int, float]], np.ndarray],
    ):
        # Un-normalise the features by reversing the min-max feature scaling
        # Increments through a 2D grid
        x, y = cuda.grid(2)
        if x < X_norm.shape[0] and y < X_norm.shape[1]:
            X[x, y] = (X_norm[x][y] * (x_max[y] - x_min[y])) + x_min[y]

    def X_standardisation_cuda_run(
        self,
        X: np.ndarray,
        X_mean: Union[List[Union[int, float]], np.ndarray],
        X_stddev: Union[List[Union[int, float]], np.ndarray],
    ) -> np.ndarray:
        X_standardised = np.zeros_like(X)
        self.numba_cuda_2d_kernel(X)
        self.X_standardisation_cuda[self.blockspergrid2d, self.threadsperblock2d](X, X_standardised, X_mean, X_stddev)

        return X_standardised

    @staticmethod
    @cuda.jit
    def X_standardisation_cuda(
        X: np.ndarray,
        X_standardised: np.ndarray,
        X_mean: Union[List[Union[int, float]], np.ndarray],
        X_stddev: Union[List[Union[int, float]], np.ndarray],
    ) -> None:
        # Increments through a 2D grid
        x, y = cuda.grid(2)
        if x < X.shape[0] and y < X.shape[1]:
            X_standardised[x, y] = (X[x][y] - X_mean[y]) / X_stddev[y]

    def Y_standardisation_cuda_run(
        self, Y: np.ndarray, Y_mean: Union[int, float], Y_stddev: Union[int, float]
    ) -> np.ndarray:
        if len(Y.shape) > 1:
            Y = Y.flatten()

        Y_standardised = np.zeros_like(Y)
        self.numba_cuda_1d_kernel(Y)
        self.Y_standardisation_cuda[self.blockspergrid1d, self.threadsperblock1d](Y, Y_standardised, Y_mean, Y_stddev)

        return Y_standardised

    @staticmethod
    @cuda.jit
    def Y_standardisation_cuda(
        Y: np.ndarray, Y_standardised: np.ndarray, Y_mean: Union[int, float], Y_stddev: Union[int, float]
    ) -> None:
        # Increments through a 1D grid
        x = cuda.grid(1)
        if x < Y.shape[0]:
            Y_standardised[x] = (Y[x] - Y_mean) / Y_stddev

    def Y_inverse_standardisation_cuda_run(
        self, Y_standardised: np.ndarray, Y_mean: Union[int, float], Y_stddev: Union[int, float]
    ) -> np.ndarray:
        if len(Y_standardised.shape) > 1:
            Y_standardised = Y_standardised.flatten()

        Y = np.zeros_like(Y_standardised)
        self.numba_cuda_1d_kernel(Y_standardised)
        self.Y_inverse_standardisation_cuda[self.blockspergrid1d, self.threadsperblock1d](
            Y_standardised, Y, Y_mean, Y_stddev
        )

        return Y

    @staticmethod
    @cuda.jit
    def Y_inverse_standardisation_cuda(
        Y_standardised: np.ndarray, Y: np.ndarray, Y_mean: Union[int, float], Y_stddev: Union[int, float]
    ) -> None:
        # Increments through a 1D grid
        x = cuda.grid(1)
        if x < Y.shape[0]:
            Y[x] = (Y_standardised[x] * Y_stddev) + Y_mean



import os
from dataclasses import dataclass
from itertools import combinations
from typing import Optional, Tuple

import numpy as np
import torch
from botorch.utils.multi_objective.hypervolume import Hypervolume
from botorch.utils.multi_objective.pareto import is_non_dominated
from pymoo.core.result import Result
from pymoo.factory import get_reference_directions, get_termination
from pymoo.optimize import minimize
from scipy import spatial

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.acquisition_functions.base_acq_function import AcquisitionFunction
from nemo_bo.acquisition_functions.nsga_improvement.problem_wrapper import NSGAProblemWrapper
from nemo_bo.acquisition_functions.nsga_improvement.unsga3 import UNSGA3
from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList
from nemo_bo.opt.samplers import LatinHyperCubeSampling, PoolBased, SampleGenerator
from nemo_bo.opt.variables import VariablesList
from nemo_bo.utils.data_proc import remove_all_nan_rows

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass
class NSGAImprovement(AcquisitionFunction):
    """

    For single-objective and multi-objective problems, this acquisition function can select between a method that
    identifies the highest hypervolume improvement using a U-NSGA3 algorithm for selecting candidates, or an
    explorative method that identifies the approximate region of the variable space that has the highest uncertainty
    in the objective prediction

    Parameters
    ----------
    num_candidates: int
        The number of sets of X arrays to be suggested by the acquisition function

    """

    num_candidates: int

    def generate_candidates(
        self,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        sampler: Optional[SampleGenerator] = None,
        constraints: Optional[ConstraintsList] = None,
        **acq_func_kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that is called when generate the best candidates using the expected improvement or expected
        hypervolume acquisition functions. This selection is made automatically based on the number of objectives

        Parameters
        ----------
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        sampler: SampleGenerator, Default = None
            Used to generate samples (SampleGenerator). PoolBased is not supported
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest hypervolume improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest hypervolume improvement

        """
        if isinstance(sampler, PoolBased):
            raise NotImplementedError(
                "The PoolBased sampler is incompatible with the NSGAImprovement acquisition function. Please select a different sampler if you wish to continue using this acquisition function. Alternatively, you may use the ExpectedImprovement acquisition function with the PoolBased sampler instead"
            )

        self.exploit_or_explore = acq_func_kwargs.get("exploit_or_explore", "exploit")
        self.pop_size = acq_func_kwargs.get("pop_size")
        self.generations = acq_func_kwargs.get("generations")

        Y = remove_all_nan_rows(Y)

        if self.pop_size is None:
            if self.num_candidates == 1:
                self.pop_size = 200
            elif self.num_candidates == 2:
                self.pop_size = 100
            elif self.num_candidates == 3:
                self.pop_size = 70
            elif self.num_candidates == 4:
                # self.pop_size = 40
                self.pop_size = 200

        if self.generations is None:
            self.generations = 30
            # if self.num_candidates == 1:
            #     self.generations = 50
            # elif self.num_candidates == 2:
            #     self.generations = 30
            # elif self.num_candidates == 3:
            #     self.generations = 30
            # elif self.num_candidates == 4:
            #     self.generations = 30

        if self.exploit_or_explore == "exploit":
            self.generate_pareto(variables, objectives, constraints)
            selected_X, selected_Y = self.hvi(self.res.X, self.res.F, Y, variables, objectives)

            self.reset = False

        if self.exploit_or_explore == "explore" or not self.hvi_search_pass:
            selected_X, selected_Y, total_Y_std = self.highest_uncertainty(variables, objectives, sampler, constraints)

            self.exploit_or_explore = "exploit"
            self.reset = True

        return selected_X, selected_Y

    def generate_pareto(
        self, variables: VariablesList, objectives: ObjectivesList, constraints: ConstraintsList
    ) -> Result:
        """
        Runs the U-NSGA3 method to generate the samples that have the highest hypervolume

        Parameters
        ----------
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        res: Result
            Contains all of the information and results from the U-NSGA3 method

        """

        xl = 0
        xu = 1

        ref_dirs = get_reference_directions("energy", objectives.n_obj, self.pop_size * 0.9, seed=1)

        logger.info(f"Performing U-NSGA3 opt")
        algorithm = UNSGA3(ref_dirs, pop_size=self.pop_size)
        problem = NSGAProblemWrapper(xl, xu, variables, objectives, constraints)
        termination = get_termination("n_gen", self.generations)

        # Performs the U-NSGA3 optimisation
        self.res = minimize(problem, algorithm, termination, verbose=True)

        # Correct by * -1 for maximising objectives in U-NSGA3 optimisation
        for obj_index, obj in enumerate(objectives.max_bool_dict):
            if objectives.max_bool_dict[obj]:
                if objectives.n_obj > 1:
                    self.res.F[:, obj_index] *= -1
                else:
                    self.res.F[obj_index] *= -1

        # Undo min-max normalisation of X values identified by U-NSGA3
        self.res.X = (self.res.X * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))) + np.array(
            variables.lower_bounds
        )

        return self.res

    def hvi(
        self,
        X_nsga: np.ndarray,
        Y_nsga: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Identifies the candidate(s) from the solutions from the U-NSGA3 optimisation that will produce the highest
        hypervolume improvement

        Parameters
        ----------
        X_nsga: np.ndarray
            Variable values for the solutions from the U-NSGA3 optimisation
        Y_nsga : np.ndarray
            Objective values for the solutions from the U-NSGA3 optimisation
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest hypervolume improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest hypervolume improvement

        """
        logger.info(f"Identifying conditions with the largest hypervolume")

        # Construct hvrefs refs
        ref_point = self.build_ref_point(objectives.max_bool_dict)

        # min-max normalisation for datasets
        sign_adjusted_Y = self.Y_norm_minmax_transform(Y, objectives.bounds, objectives.max_bool_dict)
        sign_adjusted_Y_nsga = self.Y_norm_minmax_transform(Y_nsga, objectives.bounds, objectives.max_bool_dict)

        # Creates a list of hypervolumes after each candidate(s) that successfully improves the hypervolume
        hv_list = []
        candidates_X_list = []
        candidates_Y_list = []
        X_nsga_combos = list(combinations(X_nsga, self.num_candidates))
        Y_nsga_combos = list(combinations(Y_nsga, self.num_candidates))
        sign_adjusted_Y_nsga_combos = list(combinations(sign_adjusted_Y_nsga, self.num_candidates))
        for x, y, candidates_Y in zip(X_nsga_combos, Y_nsga_combos, sign_adjusted_Y_nsga_combos):
            # Adds the candidate objective values to the experimental data set
            YappF = np.vstack([sign_adjusted_Y, np.array(candidates_Y)])
            # Identifies which data points are non-dominated
            pareto_mask_YappF = is_non_dominated(torch.tensor(YappF, dtype=torch.double))
            # Creates an numpy.ndarray containing the non-dominated points
            pareto_front_YappF = YappF[pareto_mask_YappF]

            # The following are checks to confirm that all new points in the candidates_Y set are on the pareto front
            candidate_in_pareto_counter = 0
            for a in candidates_Y:
                if any((a == b).all() for b in pareto_front_YappF):
                    candidate_in_pareto_counter += 1
                else:
                    break
            # So if all new points in the candidates_Y set are on the pareto front, then it will append the x and y and calculate the predicted hypervolume
            if candidate_in_pareto_counter == len(candidates_Y):
                candidates_X_list.append(x)
                candidates_Y_list.append(y)
                # Calculates the hypervolume
                hv = Hypervolume(ref_point=torch.tensor(ref_point, dtype=torch.double))
                hv_list.append(hv.compute(torch.tensor(pareto_front_YappF, dtype=torch.double)))

        # Selects the variable and objective values of the candidate(s) that produced the largest hypervolume
        # if hv_list is not empty
        if hv_list:
            logger.info("Conditions were successfully selected using hypervolume improvement")
            max_idx = hv_list.index(max(hv_list))
            selected_X = np.array(candidates_X_list[max_idx])
            selected_Y = np.array(candidates_Y_list[max_idx])
            self.hvi_search_pass = True

            if variables.num_cat_descriptor_var > 0:
                selected_X = variables.descriptor_to_name(selected_X)

        else:
            logger.info(
                "Failed to identify a set of conditions that would improve the hypervolume at the desired batch size"
            )
            selected_X = np.array([0])
            selected_Y = np.array([0])
            self.hvi_search_pass = False

        return selected_X, selected_Y

    def highest_uncertainty(
        self,
        variables: VariablesList,
        objectives: ObjectivesList,
        sampler: Optional[SampleGenerator] = None,
        constraints: Optional[ConstraintsList] = None,
        num_new_points: int = 2000000,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """

        Function to identify a set of input variables from the variable space that has the highest uncertainty for
        predicted output value(s)

        Parameters
        ----------
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        sampler: SampleGenerator, Default = None
            Used to generate samples (SampleGenerator). PoolBased is not supported
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints
        num_new_points: int, Default: 200,000
            Number of new sets of input variables to generate

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest hypervolume improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest hypervolume improvement
        total_Y_std: np.ndarray
            Array of total min-max normalised standard deviations for Y-values of the identified candidates' that have
            the highest hypervolume improvement

        """
        logger.info(f"Identifying conditions with the highest uncertainty")

        if sampler is None:
            sampler = LatinHyperCubeSampling(num_new_points)
        else:
            # if sampler.num_new_points is None:
            # sampler.num_new_points = num_new_points
            sampler.num_new_points = num_new_points
            # sampler = sampler

        X_new_points = sampler.generate_samples(variables, constraints)

        Y_new_points, Y_new_points_stddev = objectives.evaluate(X_new_points)

        # Min_max normalisation of Y standard deviation
        # I think this should be fine to minmax normalise like this
        for obj_index, _ in enumerate(objectives.bounds):
            Y_new_points_stddev[:, obj_index] = Y_new_points_stddev[:, obj_index] / (
                objectives.bounds[obj_index, 1] - objectives.bounds[obj_index, 0]
            )

        Y_new_points_stddev_sum = np.sum(Y_new_points_stddev, axis=1).reshape(-1, 1)
        sorting_indices = Y_new_points_stddev_sum[:, -1].argsort()

        X_new_points_sorted = X_new_points[sorting_indices][::-1]

        if variables.num_cat_var > 0:
            X_new_points = variables.categorical_transform(X_new_points).astype("float")
        X_new_points_norm_sorted = (
            (X_new_points[sorting_indices][::-1] - np.array(variables.lower_bounds))
            / (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
        )[sorting_indices][::-1]
        Y_new_points_sorted = Y_new_points[sorting_indices][::-1]
        Y_new_points_stddev_sum_sorted = Y_new_points_stddev_sum[sorting_indices][::-1]

        # 0 index for the highest uncertainty one at the top of the array
        chosen_indices = [0]
        # cosine similarity index of 1.0 for the highest uncertainty one at the top of the array
        cosine_simularity_list = [1.0]
        # Comparing the X_values for the highest uncertainty with every other entry down the sorted X_array
        for index, x in enumerate(X_new_points_norm_sorted):
            cosine_simularity_index = 1 - spatial.distance.cosine(X_new_points_norm_sorted[0], x)
            # Only passes if the cosine similarity index of a given x row is greater than 0.1 different from all of the x's in the list
            if all((((x - cosine_simularity_index) ** 2) ** 0.5) > 0.1 for x in cosine_simularity_list):
                cosine_simularity_list.append(cosine_simularity_index)
                chosen_indices.append(index)

            if len(chosen_indices) == self.num_candidates:
                break

        selected_X = X_new_points_sorted[chosen_indices]
        selected_Y = Y_new_points_sorted[chosen_indices].astype(np.float64)
        total_Y_std = Y_new_points_stddev_sum_sorted[chosen_indices].astype(np.float64)

        logger.info(f"Identified variables with the most uncertain output values")

        return selected_X, selected_Y, total_Y_std



import ast
import os
import pickle
from attrs import define, field
from datetime import datetime
from typing import Any, Dict, Optional, Tuple, List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from botorch.utils.multi_objective.hypervolume import Hypervolume
from botorch.utils.multi_objective.pareto import is_non_dominated
from gpytorch.kernels import AdditiveStructureKernel

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.plotter as plotter
from nemo_bo.acquisition_functions.base_acq_function import AcquisitionFunction
from nemo_bo.acquisition_functions.expected_improvement.expected_improvement import ExpectedImprovement
from nemo_bo.opt.benchmark import Benchmark, ModelBenchmark, SyntheticBenchmark
from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList, RegressionObjective
from nemo_bo.opt.samplers import LatinHyperCubeSampling, PoolBased, SampleGenerator
from nemo_bo.opt.variables import VariablesList
from nemo_bo.utils.data_proc import remove_nan, remove_all_nan_rows

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@define
class Optimisation:
    """

    Class used to optimise a problem defined by the variables, objectives, acquisition function used to suggest new
    experiments, and optional constraints, sampler, and/or benchmark functions

    Parameters
    ----------
    variables: VariablesList
        VariablesList object that contains information about all variables
    objectives: ObjectivesList
        ObjectivesList object that contains information about all objectives
    acquisition_func: AcquisitionFunction
        Specifies the method used to suggest new candidates
    sampler: SampleGenerator, PoolBased, Default = None
        Specifies the method to obtain samples, either using a generator or from a sample pool
    constraints: ConstraintsList, Default = None
        ConstraintsList object that contains information about all constraints
    benchmark_func: Benchmark, Default = None
        Provides a method to predict the outcome of the suggested candidates made by the acquisition function when a
        simulated optimisation is being performed

    """

    variables: VariablesList
    objectives: ObjectivesList
    acquisition_func: AcquisitionFunction
    sampler: Optional[Union[SampleGenerator, PoolBased]] = None
    constraints: Optional[ConstraintsList] = None
    benchmark_func: Optional[Benchmark] = None
    optimisation_dict: Dict[str, Any] = field(init=False)
    X_columns: List[str] = field(init=False)
    Y_columns: List[str] = field(init=False)
    X: np.ndarray = field(init=False)
    Y: np.ndarray = field(init=False)
    optimisation_only: pd.DataFrame = field(init=False)
    selected_X: np.ndarray = field(init=False)
    selected_Y: np.ndarray = field(init=False)

    def __attrs_post_init__(self):
        if isinstance(self.acquisition_func, ExpectedImprovement) and self.sampler is None:
            self.sampler = LatinHyperCubeSampling(num_new_points=1024)

        # Dictionary that stores the optimisation information for every iteration
        self.optimisation_dict = {}

        # Constructs lists of variables and objectives with their respective units where provided
        self.X_columns = []
        for name, units in zip(self.variables.names, self.variables.units):
            if units == "":
                self.X_columns.append(name)
            else:
                self.X_columns.append(f"{name} ({units})")

        self.Y_columns = []
        for name, units in zip(self.objectives.names, self.objectives.units):
            if units == "":
                self.Y_columns.append(name)
            else:
                self.Y_columns.append(f"{name} ({units})")

    def run(
        self,
        X: Optional[np.ndarray] = None,
        Y: Optional[np.ndarray] = None,
        number_of_iterations: int = 200,
        test_ratio: float = 0.2,
        number_opt_iterations_completed: int = 0,
        model_run_counter: int = 0,
        model_run_counter_threshold: int = 5,
        plot_pareto_hypervolume: bool = False,
        test_data: Optional[List[np.ndarray]] = None,
        **kwargs,
    ) -> Dict[str, Dict[str, Any]]:
        """

        A convenient function to start a Bayesian optimisation, including model fitting. It will also request the results
        at the end of each iteration and save all optimisation data, and then automatically start the next iteration.

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray
            Y array that contains the unprocessed objective data for all calculable and regression objectives, i.e.
            The data has not yet undergone any transformations related to standardisation/normalisation
        number_of_iterations: int, Default = 200,
            Number of optimisation iterations to perform
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split for the validation and test sets where applicable
        number_opt_iterations_completed: int, Default = 0,
            The number of optimisation iterations completed so far. This is used to identify the training data in the
            provided X and Y arrays
        model_run_counter: int, Default = 0,
            Assigns the run counter for the first iteration when this function is called. When the model_run_counter is
            equal to model_run_counter_threshold, an automated model and hyperparameter optimisation will occur at the
            next opportunity
        model_run_counter_threshold: int, Default = 5,
            Sets the threshold value for when an automated model and hyperparameter optimisation should occur
        plot_pareto_hypervolume: bool, Default = False,
            Sets whether to generate the 2D pareto plot (2 objectives) and hypervolume over iterations plot every
            iteration
        test_data: List[np.ndarray], Default = None,
            Test data to use for the initial model and hyperparameter optimisation, instead of creating it
            automatically. The X array is at index 0 and the Y array is at index 1

        Returns
        -------
        self.optimisation_dict: Dict[str, Dict[str, Any]]
            Dictionary that stores the optimisation information for every iteration

        """
        if isinstance(self.benchmark_func, SyntheticBenchmark):
            # X = self.sampler.__class__((2 * self.variables.n_var) + 2 if self.variables.n_var >= 4 else 10).generate_samples(self.variables, self.constraints)
            X = self.sampler.__class__(
                (5 * self.variables.n_var) if self.variables.n_var >= 2 else 10
            ).generate_samples(self.variables, self.constraints)
            Y, _ = self.benchmark_func.evaluate(X)
            Y = Y.reshape(-1, 1)

        self.X = X
        self.Y = Y

        for iteration in range(number_of_iterations):
            if model_run_counter % model_run_counter_threshold == 0:
                model_search_bool = True
                print("Will perform automated model and hyperparameter selection")
            else:
                model_search_bool = False
                print("Depending on predictor type, will perform model fitting with stored hyperparameters")

            if plot_pareto_hypervolume:
                # Scatter plot of the Y data
                self.plot_scatter_opt(
                    Y=self.Y, number_opt_iterations=number_opt_iterations_completed,
                )

                # Plot hypervolume over iterations plot
                self.plot_hypervolume(
                    Y=self.Y, number_opt_iterations=number_opt_iterations_completed,
                )

            # Obtain suggested candidates
            self.selected_X, self.selected_Y = self.find_candidates(
                self.X,
                self.Y,
                model_search_bool=model_search_bool,
                test_ratio=test_ratio,
                test_data=test_data,
                **kwargs,
            )
            if self.selected_Y.ndim == 1:
                self.selected_Y = self.selected_Y.reshape(-1, 1)

            # Creates a DataFrame of the suggested candidates
            suggested_XY = pd.DataFrame(
                np.hstack((self.selected_X, self.selected_Y)), columns=self.X_columns + self.Y_columns,
            )
            with pd.option_context("display.max_rows", None, "display.max_columns", None):
                print(suggested_XY)

            # If the optimisation is not a simulated one that used a sample pool or a benchmark function
            if not isinstance(self.sampler, PoolBased) and self.benchmark_func is None:
                if not os.path.exists(os.path.join(os.getcwd(), "Results")):
                    os.makedirs(os.path.join(os.getcwd(), "Results"))
                # Creates an Excel file containing the suggested candidates
                suggested_XY.to_excel(
                    os.path.join(
                        os.getcwd(),
                        "Results",
                        f"Iteration {iteration + 1} suggested candidates {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.xlsx",
                    )
                )
                while True:
                    try:
                        # Requests a new array containing the actual Y-values. The array has to be the correct shape
                        # and a list type to be accepted
                        actual_result = ast.literal_eval(
                            input(
                                f"Variables for the current opt iteration have been suggested (see print or 'Iteration {iteration + 1} suggested candidates.xlsx')\nPlease input the actual results in a python list:\n"
                            )
                        )
                        assert isinstance(actual_result, list)

                        self.selected_Y = (
                            np.array(actual_result)
                            if self.acquisition_func.num_candidates > 1
                            else np.array(actual_result).reshape(1, -1)
                        )

                        assert (
                            self.selected_Y.shape[0] == self.acquisition_func.num_candidates
                            and self.selected_Y.shape[1] == self.objectives.n_obj
                            and np.issubdtype(self.selected_Y.dtype, np.number)
                        )

                        break
                    except AssertionError or IndexError or ValueError or SyntaxError:
                        print(
                            "\nAssertionError: Please ensure that you have inputted the actual numerical results in a python list with the correct dimensions for the number of candidates (x) and number of objectives (y)\n"
                        )

            # Updates the X and Y arrays with the new candidates
            self.X = np.vstack((self.X, self.selected_X))
            self.Y = np.vstack((self.Y, self.selected_Y))

            # Saves the optimisation and model information for the current iteration
            self.save(iteration, model_search_bool, model_run_counter)

            # After every iteration, adds 1 to the model_run_counter number
            if self.acquisition_func.reset or model_search_bool:
                model_run_counter = 1
            else:
                model_run_counter += 1

            number_opt_iterations_completed += 1

        return self.optimisation_dict

    def find_candidates(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        model_search_bool,
        test_ratio: float = 0.2,
        test_data: Optional[List[np.ndarray]] = None,
        **kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that returns the suggested candidates by:
        1. Fitting any required regression models
        2. Uses an acquisition function to suggest the next best candidates
        3. Updates the Y-values with

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data for all calculable and regression objectives, i.e.
            The data has not yet undergone any transformations related to standardisation/normalisation
        model_search_bool: bool
            Whether automated model and hyperparameter optimisation is to be performed when fitting the regression
            models
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split for the validation and test sets where applicable
        test_data: List[np.ndarray], Default = None,
            Test data to use for the initial model and hyperparameter optimisation, instead of creating it
            automatically. The X array is at index 0 and the Y array is at index 1

        Returns
        -------
        selected_X: np.ndarray
            The X array of the suggested candidates
        selected_Y: np.ndarray
            The Y array of the suggested candidates

        """
        if any(isinstance(obj, RegressionObjective) for obj in self.objectives.objectives):

            # Fit the ML regression models
            self.objectives.fit(
                X, Y, self.variables, model_search_bool, test_ratio=test_ratio, test_data=test_data, **kwargs
            )

        # Use the fitted regression models, calculable objectives, and acquisition function to suggest new candidates
        selected_X, selected_Y = self.acquisition_func.generate_candidates(
            Y, self.variables, self.objectives, self.sampler, self.constraints, **kwargs
        )

        # If a sample pool was used, obtain the actual Y-values and removing the chosen samples from the sample pool
        if isinstance(self.sampler, PoolBased):
            selected_Y = self.sampler.Y_and_update_pool(self.sampler.index)

        else:
            # If a benchmark function is provided, use it to predict the outcome of the suggested X array
            if self.benchmark_func is not None:
                if self.benchmark_func.fitted:
                    selected_Y, _ = self.benchmark_func.evaluate(selected_X)
                else:
                    raise AttributeError("Please firstly fit the benchmark function(s) and check its quality")

        return selected_X, selected_Y

    def save(self, iteration: int, model_search_bool: bool, model_run_counter: int) -> None:
        """

        Saves the optimisation and model information for the current iteration

        Parameters
        ----------
        iteration: int
            Iteration number for the current optimisation
        model_search_bool: bool
            Whether automated model and hyperparameter optimisation was performed when the regression models were fitted
        model_run_counter: int
            The run counter for this optimisation iteration

        """
        # Save excel for current iteration data
        if not os.path.exists(os.path.join(os.getcwd(), "Results")):
            os.makedirs(os.path.join(os.getcwd(), "Results"))

        # Temporarily stored the information in these lists
        model_info_list = [model_search_bool, model_run_counter]
        model_info_columns = [
            "Model and Hyperparameter Optimisation",
            "Model Run Counter",
        ]
        for obj in self.objectives.objectives:
            if isinstance(obj, RegressionObjective):
                if obj.obj_function.include_validation:
                    model_info_list.append(obj.obj_function.performance_metrics["Train RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["Train r2"])
                    model_info_list.append(obj.obj_function.performance_metrics["Validation RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["Validation r2"])
                    model_info_columns.append(f"{obj.name} Model Train RMSE")
                    model_info_columns.append(f"{obj.name} Model Train r2")
                    model_info_columns.append(f"{obj.name} Model Validation RMSE")
                    model_info_columns.append(f"{obj.name} Model Validation r2")
                else:
                    model_info_list.append(obj.obj_function.performance_metrics["RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["r2"])
                    model_info_columns.append(f"{obj.name} Model Train RMSE")
                    model_info_columns.append(f"{obj.name} Model Train r2")
                model_info_list.append(obj.obj_function.name)

                if obj.obj_function.name == "gp":
                    if isinstance(obj.model_params["kernel"], AdditiveStructureKernel):
                        model_info_list.append(
                            f"AdditiveStructureKernel(ScaleKernel(MaternKernel(ard_num_dims=n_var, nu={obj.model_params['kernel'].base_kernel.base_kernel.nu}, lengthscale_prior=GammaPrior(3.0, 6.0)), outputscale_prior=GammaPrior(2.0, 0.15)), num_dims=n_var)"
                        )
                    else:
                        model_info_list.append(
                            f"ScaleKernel(MaternKernel(nu={obj.model_params['kernel'].base_kernel.nu}, ard_num_dims=n_var, lengthscale_prior=GammaPrior(3.0, 6.0)), outputscale_prior=GammaPrior(2.0, 0.15))"
                        )
                else:
                    model_info_list.append(obj.model_params)

                model_info_columns.append(f"{obj.name} Model Type")
                model_info_columns.append(f"{obj.name} Model Hyperparameters")

        model_info_list.append(datetime.now().strftime("%d-%m-%Y, %H-%M-%S"))
        model_info_columns.append("Date and Time")
        # Duplicates this information for every candidate
        model_info = np.array([model_info_list] * self.acquisition_func.num_candidates)

        # Creates the DataFrames of optimisation information
        optimisation_results = pd.DataFrame(np.hstack((self.X, self.Y)), columns=self.X_columns + self.Y_columns)
        if iteration == 0:
            self.optimisation_only = pd.DataFrame(
                np.hstack((self.selected_X, self.selected_Y, model_info)),
                columns=self.X_columns + self.Y_columns + model_info_columns,
            )
        else:
            self.optimisation_only = pd.concat(
                (
                    self.optimisation_only,
                    pd.DataFrame(
                        np.hstack((self.selected_X, self.selected_Y, model_info)),
                        columns=self.X_columns + self.Y_columns + model_info_columns,
                    ),
                ),
                axis=0,
                ignore_index=True,
            )

        # Writes the optimisation information DataFrames into Excel files
        with pd.ExcelWriter(
            os.path.join(
                os.getcwd(),
                "Results",
                f'Optimisation Results (after {iteration + 1} iteration(s)) {datetime.now().strftime("%d-%m-%Y, %H-%M-%S")}.xlsx',
            )
        ) as writer:
            optimisation_results.to_excel(writer, sheet_name="Training + Opt")
            self.optimisation_only.to_excel(writer, sheet_name="Only Opt")
            for obj in self.objectives.objectives:
                if isinstance(obj, RegressionObjective):
                    if obj.obj_function.include_validation:
                        obj.obj_function.performance_metrics["Train Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Train Parity Data"
                        )
                        obj.obj_function.performance_metrics["Validation Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Validation Parity Data"
                        )
                    else:
                        obj.obj_function.performance_metrics["Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Train Parity Data"
                        )

        # Writes pickle files containing every iteration's optimisation information
        self.optimisation_dict[f"Iteration {iteration + 1}"] = {}
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Date and Time"] = datetime.now().strftime(
            "%d-%m-%Y, %H-%M-%S"
        )
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Training + Optimisation Table"] = optimisation_results
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Optimisation Table"] = self.optimisation_only
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Optimisation Selected X"] = self.selected_X
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Optimisation Selected Y"] = self.selected_Y
        self.optimisation_dict[f"Iteration {iteration + 1}"][
            "Model and Hyperparameter Optimisation"
        ] = model_search_bool
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Model Run Counter"] = model_run_counter
        for obj in self.objectives.objectives:
            if isinstance(obj, RegressionObjective):
                self.optimisation_dict[f"Iteration {iteration + 1}"][f"{obj.name} Model Type"] = obj.obj_function.name
                if obj.obj_function.name == "gp":
                    if isinstance(obj.model_params["kernel"], AdditiveStructureKernel):
                        self.optimisation_dict[f"Iteration {iteration + 1}"][
                            f"{obj.name} Model Hyperparameters"
                        ] = f"AdditiveStructureKernel(ScaleKernel(MaternKernel(ard_num_dims=n_var, nu={obj.model_params['kernel'].base_kernel.base_kernel.nu}, lengthscale_prior=GammaPrior(3.0, 6.0)), outputscale_prior=GammaPrior(2.0, 0.15)), num_dims=n_var)"
                    else:
                        self.optimisation_dict[f"Iteration {iteration + 1}"][
                            f"{obj.name} Model Hyperparameters"
                        ] = f"ScaleKernel(MaternKernel(nu={obj.model_params['kernel'].base_kernel.nu}, ard_num_dims=n_var, lengthscale_prior=GammaPrior(3.0, 6.0)), outputscale_prior=GammaPrior(2.0, 0.15))"
                else:
                    self.optimisation_dict[f"Iteration {iteration + 1}"][
                        f"{obj.name} Model Hyperparameters"
                    ] = obj.model_params
                self.optimisation_dict[f"Iteration {iteration + 1}"][
                    f"{obj.name} Model Performance Metrics"
                ] = obj.obj_function.performance_metrics

        with open(
            os.path.join(
                os.getcwd(),
                "Results",
                f"Optimisation Results (after {iteration + 1} iteration(s)) {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.pickle",
            ),
            "wb",
        ) as handle:
            pickle.dump(self.optimisation_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def calc_hypervolume(self, Y: np.ndarray) -> float:
        """

        Calculates the hypervolume of a Y-array that has been min-max normalised by the pre-defined objective bounds

        Parameters
        ----------
        Y: np.ndarray
            Y-array with objective data. Any data points with NaN values will be removed

        Returns
        -------
        The normalised hypervolume of a Y-array

        """
        Y_no_nan = remove_all_nan_rows(Y)
        ref_point = self.acquisition_func.build_ref_point(self.objectives.max_bool_dict)
        sign_adjusted_Y = self.acquisition_func.Y_norm_minmax_transform(
            Y_no_nan, self.objectives.bounds, self.objectives.max_bool_dict
        )

        # Calculates pareto front
        pareto_mask = is_non_dominated(torch.tensor(sign_adjusted_Y, dtype=torch.double))
        pareto_front = sign_adjusted_Y[pareto_mask]

        # Calculates the hypervolume
        hv = Hypervolume(ref_point=torch.tensor(ref_point, dtype=torch.double))
        return hv.compute(torch.tensor(pareto_front, dtype=torch.double))

    def plot_scatter_opt(self, Y: np.ndarray, number_opt_iterations: int) -> None:
        """

        Plots the Y-array for 1, 2, or 3 objective optimisation problems, with multi-objective pareto fronts when
        applicable

        Parameters
        ----------
        Y: np.ndarray
            Y-array with objective data. Any data points with NaN values will be removed
        number_opt_iterations: int
            Number of Bayesian optimisation iterations performed so far

        """
        train_num = Y.shape[0] - (number_opt_iterations * self.acquisition_func.num_candidates)
        opt_num = number_opt_iterations * self.acquisition_func.num_candidates

        if self.objectives.n_obj == 1:
            iterations = np.array((range(Y.shape[0]))) + 1
            Y = np.hstack((iterations.reshape(-1, 1), Y))

        else:
            Y_no_nan = remove_all_nan_rows(Y)

            sign_adjusted_Y = self.acquisition_func.Y_norm_minmax_transform(
                Y_no_nan, self.objectives.bounds, self.objectives.max_bool_dict
            )

            # Calculates pareto front
            Y_to_expnum = torch.tensor(np.array(sign_adjusted_Y), dtype=torch.double)
            pareto_mask_to_expnum = is_non_dominated(Y_to_expnum)
            pareto_front = Y_no_nan[pareto_mask_to_expnum]

            non_zero_objs = np.where(np.all(~np.isclose(Y_no_nan, 0.0), axis=0))[0]
            if len(non_zero_objs) > 0:
                # Uses the first objective in the array that has all non-zero values
                pareto_front_sorted = pareto_front[pareto_front[:, non_zero_objs[0]].argsort()]
            else:
                # Else just use the first objective
                pareto_front_sorted = pareto_front[pareto_front[:, 0].argsort()]

        train_experiments = Y[
            :train_num, :,
        ]
        opt_experiments = Y[
            train_num:, :,
        ]

        train_experiments = remove_all_nan_rows(train_experiments)
        opt_experiments = remove_all_nan_rows(opt_experiments)

        scatter_data = [train_experiments, opt_experiments]
        scatter_legend = [
            f"Training ({train_num} experiments)"
            if train_num == train_experiments.shape[0]
            else f"Training ({train_num} expt, showing {train_experiments.shape[0]} expt)",
            f"Optimisation ({opt_num} experiments)"
            if train_num == opt_experiments.shape[0]
            else f"Optimisation ({opt_num} expt, showing {opt_experiments.shape[0]} expt)",
        ]

        title = "Pareto Plot"
        output = os.path.join(os.getcwd(), f'Pareto Plot {datetime.now().strftime("%d-%m-%Y, %H-%M-%S")}.png',)

        if self.objectives.n_obj == 1:
            plotter.plot(
                plot_dim="2D",
                scatter_data=scatter_data,
                # Create array to use for drawing the start of the optimisation line
                line_data=[[train_num + 0.5, min(Y[:, 1])], [train_num + 0.5, max(Y[:, 1])]],
                scatter_legend=scatter_legend,
                line_legend="Start of Optimisation",
                xlabel=f"Number of Iterations",
                ylabel=f"{self.objectives.names[0]} {self.objectives.units[0]}",
                plottitle=f"{title}\n{self.objectives.names[0]} vs. Number of Iterations",
                output_file=rf"{output}",
            )
        elif self.objectives.n_obj == 2:
            plotter.plot(
                plot_dim="2D",
                scatter_data=scatter_data,
                line_data=pareto_front_sorted,
                scatter_legend=scatter_legend,
                line_legend="Pareto front",
                xlabel=f"{self.objectives.names[0]} {self.objectives.units[0]}",
                ylabel=f"{self.objectives.names[1]} {self.objectives.units[1]}",
                plottitle=f"{title}\n{self.objectives.names[0]} vs. {self.objectives.names[1]}",
                output_file=rf"{output}",
            )
        elif self.objectives.n_obj == 3:
            plotter.plot(
                plot_dim="3D",
                scatter_data=scatter_data,
                surface_data=pareto_front_sorted,
                scatter_legend=scatter_legend,
                surface_legend="Pareto front",
                xlabel=f"{self.objectives.names[0]} {self.objectives.units[0]}",
                ylabel=f"{self.objectives.names[1]} {self.objectives.units[1]}",
                zlabel=f"{self.objectives.names[2]} {self.objectives.units[2]}",
                plottitle=f"{title}\n{self.objectives.names[0]} vs. {self.objectives.names[1]} vs. {self.objectives.names[2]}",
                output_file=rf"{output}",
            )

        plt.close()

    def plot_hypervolume(self, Y: np.ndarray, number_opt_iterations: int) -> None:
        """

        Plots the hypervolume over all iterations

        Parameters
        ----------
        Y: np.ndarray
            Y-array with objective data. Any data points with NaN values will be removed
        number_opt_iterations: int
            Number of Bayesian optimisation iterations performed so far

        """
        train_num = Y.shape[0] - (number_opt_iterations * self.acquisition_func.num_candidates)
        opt_num = number_opt_iterations * self.acquisition_func.num_candidates

        iterations = np.array((range(Y.shape[0]))) + 1
        del_index = [index for index in range(Y.shape[0]) if np.any(np.isnan(Y[index]), axis=0)]
        iterations = np.delete(iterations, del_index, axis=0)

        no_nan_train_num = sum(expt_number <= train_num for expt_number in iterations)
        no_nan_opt_num = sum(expt_number > train_num for expt_number in iterations)

        Y_no_nan = remove_all_nan_rows(Y)
        ref_point = self.acquisition_func.build_ref_point(self.objectives.max_bool_dict)
        sign_adjusted_Y = self.acquisition_func.Y_norm_minmax_transform(
            Y_no_nan, self.objectives.bounds, self.objectives.max_bool_dict
        )
        # Recalculates the hypervolume for every additional row in the data
        hv_list = [0]
        for y_index in range(2, Y_no_nan.shape[0] + 1):
            # Calculates pareto front
            obj_data_to_expnum = sign_adjusted_Y[:y_index, :]
            obj_data_to_expnum_np = np.array(obj_data_to_expnum)
            Y_to_expnum = torch.tensor(obj_data_to_expnum_np, dtype=torch.double)
            pareto_mask_to_expnum = is_non_dominated(Y_to_expnum)
            pareto_front_to_expnum = obj_data_to_expnum_np[pareto_mask_to_expnum]

            # Calculates the hypervolume
            hv = Hypervolume(ref_point=torch.tensor(ref_point, dtype=torch.double))
            hv_list.append(hv.compute(torch.tensor(pareto_front_to_expnum, dtype=torch.double)))

        if Y.shape[0] == len(iterations):
            title = f"Hypervolume\nTraining: {train_num} expt\nOptimisation: {opt_num} expt"
        else:
            title = f"Hypervolume\nTraining: {train_num} expt, showing {no_nan_train_num} expt\nOptimisation: {opt_num} expt, showing {no_nan_opt_num} expt"

        output_file = os.path.join(
            os.getcwd(), rf"Hypervolume vs. Iteration {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.png",
        )

        hv_array = np.array([[expt_number, hv_list[index]] for index, expt_number in enumerate(iterations)])
        plotter.plot(
            plot_dim="2D",
            scatter_data=hv_array,
            line_data=[[train_num + 0.5, 0.0], [train_num + 0.5, 1.0]],
            scatter_legend="Hypervolume",
            line_legend="Start of Optimisation",
            xlabel=f"Iteration",
            ylabel=f"Normalised Hypervolume",
            plottitle=title,
            output_file=output_file,
        )
        plt.close()
